@@ -3,12 +3,6 @@ import Cocoa
 import SwiftUI
 import GhosttyKit
 
-// This is a Apple's private function that we need to call to get the active space.
-@_silgen_name("CGSGetActiveSpace")
-func CGSGetActiveSpace(_ cid: Int) -> size_t
-@_silgen_name("CGSMainConnectionID")
-func CGSMainConnectionID() -> Int
-
 /// Controller for the "quick" terminal.
 class QuickTerminalController: BaseTerminalController {
     override var windowNibName: NSNib.Name? { "QuickTerminal" }
@@ -25,22 +19,37 @@ class QuickTerminalController: BaseTerminalController {
     private var previousApp: NSRunningApplication? = nil
 
     // The active space when the quick terminal was last shown.
-    private var previousActiveSpace: size_t = 0
+    private var previousActiveSpace: CGSSpace? = nil
+
+    /// The window frame saved when the quick terminal's surface tree becomes empty.
+    ///
+    /// This preserves the user's window size and position when all terminal surfaces
+    /// are closed (e.g., via the `exit` command). When a new surface is created,
+    /// the window will be restored to this frame, preventing SwiftUI from resetting
+    /// the window to its default minimum size.
+    private var lastClosedFrame: NSRect? = nil
 
     /// Non-nil if we have hidden dock state.
     private var hiddenDock: HiddenDock? = nil
 
     /// The configuration derived from the Ghostty config so we don't need to rely on references.
     private var derivedConfig: DerivedConfig
+    
+    /// Tracks if we're currently handling a manual resize to prevent recursion
+    private var isHandlingResize: Bool = false
 
     init(_ ghostty: Ghostty.App,
          position: QuickTerminalPosition = .top,
          baseConfig base: Ghostty.SurfaceConfiguration? = nil,
-         surfaceTree tree: Ghostty.SplitNode? = nil
+         surfaceTree tree: SplitTree<Ghostty.SurfaceView>? = nil
     ) {
         self.position = position
         self.derivedConfig = DerivedConfig(ghostty.config)
-        super.init(ghostty, baseConfig: base, surfaceTree: tree)
+
+        // Important detail here: we initialize with an empty surface tree so
+        // that we don't start a terminal process. This gets started when the
+        // first terminal is shown in `animateIn`.
+        super.init(ghostty, baseConfig: base, surfaceTree: .init())
 
         // Setup our notifications for behaviors
         let center = NotificationCenter.default
@@ -51,7 +60,7 @@ class QuickTerminalController: BaseTerminalController {
             object: nil)
         center.addObserver(
             self,
-            selector: #selector(onToggleFullscreen),
+            selector: #selector(onToggleFullscreen(notification:)),
             name: Ghostty.Notification.ghosttyToggleFullscreen,
             object: nil)
         center.addObserver(
@@ -59,6 +68,22 @@ class QuickTerminalController: BaseTerminalController {
             selector: #selector(ghosttyConfigDidChange(_:)),
             name: .ghosttyConfigDidChange,
             object: nil)
+        center.addObserver(
+            self,
+            selector: #selector(closeWindow(_:)),
+            name: .ghosttyCloseWindow,
+            object: nil
+        )
+        center.addObserver(
+            self,
+            selector: #selector(onNewTab),
+            name: Ghostty.Notification.ghosttyNewTab,
+            object: nil)
+        center.addObserver(
+            self,
+            selector: #selector(windowDidResize(_:)),
+            name: NSWindow.didResizeNotification,
+            object: nil)
     }
 
     required init?(coder: NSCoder) {
@@ -92,14 +117,29 @@ class QuickTerminalController: BaseTerminalController {
         syncAppearance()
 
         // Setup our initial size based on our configured position
-        position.setLoaded(window)
+        position.setLoaded(window, size: derivedConfig.quickTerminalSize)
 
+        // Upon first adding this Window to its host view, older SwiftUI
+        // seems to have a "hiccup" and corrupts the frameRect,
+        // sometimes setting the size to zero, sometimes corrupting it.
+        // We pass the actual window's frame as "initial" frame directly
+        // to the window, so it can use that instead of the frameworks
+        // "interpretation"
+        if let qtWindow = window as? QuickTerminalWindow {
+            qtWindow.initialFrame = window.frame
+        }
+        
         // Setup our content
         window.contentView = NSHostingView(rootView: TerminalView(
             ghostty: self.ghostty,
             viewModel: self,
             delegate: self
         ))
+        
+        // Clear out our frame at this point, the fixup from above is complete.
+        if let qtWindow = window as? QuickTerminalWindow {
+            qtWindow.initialFrame = nil
+        }
 
         // Animate the window in
         animateIn()
@@ -149,14 +189,24 @@ class QuickTerminalController: BaseTerminalController {
                 animateOut()
 
             case .move:
-                let currentActiveSpace = CGSGetActiveSpace(CGSMainConnectionID())
+                let currentActiveSpace = CGSSpace.active()
                 if previousActiveSpace == currentActiveSpace {
                     // We haven't moved spaces. We lost focus to another app on the
                     // current space. Animate out.
                     animateOut()
                 } else {
-                    // We've moved to a different space. Bring the quick terminal back
-                    // into view.
+                    // We've moved to a different space.
+
+                    // If we're fullscreen, we need to exit fullscreen because the visible
+                    // bounds may have changed causing a new behavior.
+                    if let fullscreenStyle, fullscreenStyle.isFullscreen {
+                        fullscreenStyle.exit()
+                        DispatchQueue.main.async {
+                            self.onToggleFullscreen()
+                        }
+                    }
+
+                    // Make the window visible again on this space
                     DispatchQueue.main.async {
                         self.window?.makeKeyAndOrderFront(nil)
                     }
@@ -167,22 +217,77 @@ class QuickTerminalController: BaseTerminalController {
         }
     }
 
-    func windowWillResize(_ sender: NSWindow, to frameSize: NSSize) -> NSSize {
-        // We use the actual screen the window is on for this, since it should
-        // be on the proper screen.
-        guard let screen = window?.screen ?? NSScreen.main else { return frameSize }
-        return position.restrictFrameSize(frameSize, on: screen)
+    override func windowDidResize(_ notification: Notification) {
+        guard let window = notification.object as? NSWindow,
+              window == self.window,
+              visible,
+              !isHandlingResize else { return }
+        guard let screen = window.screen ?? NSScreen.main else { return }
+
+        // Prevent recursive loops
+        isHandlingResize = true
+        defer { isHandlingResize = false }
+        
+        switch position {
+        case .top, .bottom, .center:
+            // For centered positions (top, bottom, center), we need to recenter the window
+            // when it's manually resized to maintain proper positioning
+            let newOrigin = position.centeredOrigin(for: window, on: screen)
+            window.setFrameOrigin(newOrigin)
+        case .left, .right:
+            // For side positions, we may need to adjust vertical centering
+            let newOrigin = position.verticallyCenteredOrigin(for: window, on: screen)
+            window.setFrameOrigin(newOrigin)
+        }
     }
 
     // MARK: Base Controller Overrides
 
-    override func surfaceTreeDidChange(from: Ghostty.SplitNode?, to: Ghostty.SplitNode?) {
+    override func surfaceTreeDidChange(from: SplitTree<Ghostty.SurfaceView>, to: SplitTree<Ghostty.SurfaceView>) {
         super.surfaceTreeDidChange(from: from, to: to)
 
-        // If our surface tree is nil then we animate the window out.
-        if (to == nil) {
+        // If our surface tree is nil then we animate the window out. We
+        // defer reinitializing the tree to save some memory here.
+        if to.isEmpty {
             animateOut()
+            return
         }
+
+        // If we're not empty (e.g. this isn't the first set) and we're
+        // not visible, then we animate in. This allows us to show the quick
+        // terminal when things such as undo/redo are done.
+        if !from.isEmpty && !visible {
+            animateIn()
+            return
+        }
+    }
+
+    override func closeSurface(
+        _ node: SplitTree<Ghostty.SurfaceView>.Node,
+        withConfirmation: Bool = true
+    ) {
+        // If this isn't the root then we're dealing with a split closure.
+        if surfaceTree.root != node {
+            super.closeSurface(node, withConfirmation: withConfirmation)
+            return
+        }
+
+        // If this isn't a final leaf then we're dealing with a split closure
+        guard case .leaf(let surface) = node else {
+            super.closeSurface(node, withConfirmation: withConfirmation)
+            return
+        }
+
+        // If its the root, we check if the process exited. If it did,
+        // then we do empty the tree.
+        if surface.processExited {
+            surfaceTree = .init()
+            return
+        }
+
+        // If its the root then we just animate out. We never actually allow
+        // the surface to fully close.
+        animateOut()
     }
 
     // MARK: Methods
@@ -219,19 +324,20 @@ class QuickTerminalController: BaseTerminalController {
         }
 
         // Set previous active space
-        self.previousActiveSpace = CGSGetActiveSpace(CGSMainConnectionID())
+        self.previousActiveSpace = CGSSpace.active()
+
+        // If our surface tree is empty then we initialize a new terminal. The surface
+        // tree can be empty if for example we run "exit" in the terminal and force
+        // animate out.
+        if surfaceTree.isEmpty,
+           let ghostty_app = ghostty.app {
+            let view = Ghostty.SurfaceView(ghostty_app, baseConfig: nil)
+            surfaceTree = SplitTree(view: view)
+            focusedSurface = view
+        }
 
         // Animate the window in
         animateWindowIn(window: window, from: position)
-
-        // If our surface tree is nil then we initialize a new terminal. The surface
-        // tree can be nil if for example we run "eixt" in the terminal and force
-        // animate out.
-        if (surfaceTree == nil) {
-            let leaf: Ghostty.SplitNode.Leaf = .init(ghostty.app!, baseConfig: nil)
-            surfaceTree = .leaf(leaf)
-            focusedSurface = leaf.surface
-        }
     }
 
     func animateOut() {
@@ -252,9 +358,17 @@ class QuickTerminalController: BaseTerminalController {
 
     private func animateWindowIn(window: NSWindow, from position: QuickTerminalPosition) {
         guard let screen = derivedConfig.quickTerminalScreen.screen else { return }
+        
+        // Grab our last closed frame to use, and clear our state since we're animating in.
+        let lastClosedFrame = self.lastClosedFrame
+        self.lastClosedFrame = nil
 
-        // Move our window off screen to the top
-        position.setInitial(in: window, on: screen)
+        // Move our window off screen to the initial animation position.
+        position.setInitial(
+            in: window,
+            on: screen,
+            terminalSize: derivedConfig.quickTerminalSize,
+            closedFrame: lastClosedFrame)
 
         // We need to set our window level to a high value. In testing, only
         // popUpMenu and above do what we want. This gets it above the menu bar
@@ -285,7 +399,11 @@ class QuickTerminalController: BaseTerminalController {
         NSAnimationContext.runAnimationGroup({ context in
             context.duration = derivedConfig.quickTerminalAnimationDuration
             context.timingFunction = .init(name: .easeIn)
-            position.setFinal(in: window.animator(), on: screen)
+            position.setFinal(
+                in: window.animator(),
+                on: screen,
+                terminalSize: derivedConfig.quickTerminalSize,
+                closedFrame: lastClosedFrame)
         }, completionHandler: {
             // There is a very minor delay here so waiting at least an event loop tick
             // keeps us safe from the view not being on the window.
@@ -363,6 +481,20 @@ class QuickTerminalController: BaseTerminalController {
     }
 
     private func animateWindowOut(window: NSWindow, to position: QuickTerminalPosition) {
+        // If we are in fullscreen, then we exit fullscreen. We do this immediately so
+        // we have th correct window.frame for the save state below.
+        if let fullscreenStyle, fullscreenStyle.isFullscreen {
+            fullscreenStyle.exit()
+        }
+
+        // Save the current window frame before animating out. This preserves
+        // the user's preferred window size and position for when the quick
+        // terminal is reactivated with a new surface. Without this, SwiftUI
+        // would reset the window to its minimum content size.
+        if window.frame.width > 0 && window.frame.height > 0 {
+            lastClosedFrame = window.frame
+        }
+
         // If we hid the dock then we unhide it.
         hiddenDock = nil
 
@@ -377,11 +509,6 @@ class QuickTerminalController: BaseTerminalController {
         // We always animate out to whatever screen the window is actually on.
         guard let screen = window.screen ?? NSScreen.main else { return }
 
-        // If we are in fullscreen, then we exit fullscreen.
-        if let fullscreenStyle, fullscreenStyle.isFullscreen {
-            fullscreenStyle.exit()
-        }
-
         // If we have a previously active application, restore focus to it. We
         // do this BEFORE the animation below because when the animation completes
         // macOS will bring forward another window.
@@ -403,7 +530,11 @@ class QuickTerminalController: BaseTerminalController {
         NSAnimationContext.runAnimationGroup({ context in
             context.duration = derivedConfig.quickTerminalAnimationDuration
             context.timingFunction = .init(name: .easeIn)
-            position.setInitial(in: window.animator(), on: screen)
+            position.setInitial(
+                in: window.animator(),
+                on: screen,
+                terminalSize: derivedConfig.quickTerminalSize,
+                closedFrame: window.frame)
         }, completionHandler: {
             // This causes the window to be removed from the screen list and macOS
             // handles what should be focused next.
@@ -437,6 +568,16 @@ class QuickTerminalController: BaseTerminalController {
         }
     }
 
+    private func showNoNewTabAlert() {
+        guard let window else { return }
+        let alert = NSAlert()
+        alert.messageText = "Cannot Create New Tab"
+        alert.informativeText = "Tabs aren't supported in the Quick Terminal."
+        alert.addButton(withTitle: "OK")
+        alert.alertStyle = .warning
+        alert.beginSheetModal(for: window)
+    }
+
     // MARK: First Responder
 
     @IBAction override func closeWindow(_ sender: Any) {
@@ -445,13 +586,7 @@ class QuickTerminalController: BaseTerminalController {
     }
 
     @IBAction func newTab(_ sender: Any?) {
-        guard let window else { return }
-        let alert = NSAlert()
-        alert.messageText = "Cannot Create New Tab"
-        alert.informativeText = "Tabs aren't supported in the Quick Terminal."
-        alert.addButton(withTitle: "OK")
-        alert.alertStyle = .warning
-        alert.beginSheetModal(for: window)
+        showNoNewTabAlert()
     }
 
     @IBAction func toggleGhosttyFullScreen(_ sender: Any) {
@@ -476,9 +611,29 @@ class QuickTerminalController: BaseTerminalController {
     @objc private func onToggleFullscreen(notification: SwiftUI.Notification) {
         guard let target = notification.object as? Ghostty.SurfaceView else { return }
         guard target == self.focusedSurface else { return }
+        onToggleFullscreen()
+    }
 
-        // We ignore the requested mode and always use non-native for the quick terminal
-        toggleFullscreen(mode: .nonNative)
+    private func onToggleFullscreen() {
+        // We ignore the configured fullscreen style and always use non-native
+        // because the way the quick terminal works doesn't support native.
+        let mode: FullscreenMode
+        if (NSApp.isFrontmost) {
+            // If we're frontmost and we have a notch then we keep padding
+            // so all lines of the terminal are visible.
+            if (window?.screen?.hasNotch ?? false) {
+                mode = .nonNativePaddedNotch
+            } else {
+                mode = .nonNative
+            }
+        } else {
+            // An additional detail is that if the is NOT frontmost, then our
+            // NSApp.presentationOptions will not take effect so we must always
+            // do the visible menu mode since we can't get rid of the menu.
+            mode = .nonNativeVisibleMenu
+        }
+
+        toggleFullscreen(mode: mode)
     }
 
     @objc private func ghosttyConfigDidChange(_ notification: Notification) {
@@ -497,11 +652,20 @@ class QuickTerminalController: BaseTerminalController {
         syncAppearance()
     }
 
+    @objc private func onNewTab(notification: SwiftUI.Notification) {
+        guard let surfaceView = notification.object as? Ghostty.SurfaceView else { return }
+        guard let window = surfaceView.window else { return }
+        guard window.windowController is QuickTerminalController else { return }
+        // Tabs aren't supported with Quick Terminals or derivatives
+        showNoNewTabAlert()
+    }
+
     private struct DerivedConfig {
         let quickTerminalScreen: QuickTerminalScreen
         let quickTerminalAnimationDuration: Double
         let quickTerminalAutoHide: Bool
         let quickTerminalSpaceBehavior: QuickTerminalSpaceBehavior
+        let quickTerminalSize: QuickTerminalSize
         let backgroundOpacity: Double
 
         init() {
@@ -509,6 +673,7 @@ class QuickTerminalController: BaseTerminalController {
             self.quickTerminalAnimationDuration = 0.2
             self.quickTerminalAutoHide = true
             self.quickTerminalSpaceBehavior = .move
+            self.quickTerminalSize = QuickTerminalSize()
             self.backgroundOpacity = 1.0
         }
 
@@ -517,6 +682,7 @@ class QuickTerminalController: BaseTerminalController {
             self.quickTerminalAnimationDuration = config.quickTerminalAnimationDuration
             self.quickTerminalAutoHide = config.quickTerminalAutoHide
             self.quickTerminalSpaceBehavior = config.quickTerminalSpaceBehavior
+            self.quickTerminalSize = config.quickTerminalSize
             self.backgroundOpacity = config.backgroundOpacity
         }
     }
