@@ -102,11 +102,11 @@ extension Ghostty {
         /// configuration would be "quit" action.
         ///
         /// Returns nil if there is no key equivalent for the given action.
-        func keyEquivalent(for action: String) -> KeyEquivalent? {
+        func keyboardShortcut(for action: String) -> KeyboardShortcut? {
             guard let cfg = self.config else { return nil }
 
             let trigger = ghostty_config_trigger(cfg, action, UInt(action.count))
-            return Ghostty.keyEquivalent(for: trigger)
+            return Ghostty.keyboardShortcut(for: trigger)
         }
 #endif
 
@@ -116,6 +116,14 @@ extension Ghostty {
         /// details on what each means. We only add documentation if there is a strange conversion
         /// due to the embedded library and Swift.
 
+        var bellFeatures: BellFeatures {
+            guard let config = self.config else { return .init() }
+            var v: CUnsignedInt = 0
+            let key = "bell-features"
+            guard ghostty_config_get(config, &v, key, UInt(key.count)) else { return .init() }
+            return .init(rawValue: v)
+        }
+
         var initialWindow: Bool {
             guard let config = self.config else { return true }
             var v = true;
@@ -132,6 +140,15 @@ extension Ghostty {
             return v
         }
 
+        var title: String? {
+            guard let config = self.config else { return nil }
+            var v: UnsafePointer<Int8>? = nil
+            let key = "title"
+            guard ghostty_config_get(config, &v, key, UInt(key.count)) else { return nil }
+            guard let ptr = v else { return nil }
+            return String(cString: ptr)
+        }
+
         var windowSaveState: String {
             guard let config = self.config else { return "" }
             var v: UnsafePointer<Int8>? = nil
@@ -140,14 +157,14 @@ extension Ghostty {
             guard let ptr = v else { return "" }
             return String(cString: ptr)
         }
-        
+
         var windowPositionX: Int16? {
             guard let config = self.config else { return nil }
             var v: Int16 = 0
             let key = "window-position-x"
             return ghostty_config_get(config, &v, key, UInt(key.count)) ? v : nil
         }
-        
+
         var windowPositionY: Int16? {
             guard let config = self.config else { return nil }
             var v: Int16 = 0
@@ -216,6 +233,8 @@ extension Ghostty {
                     .nonNative
             case "visible-menu":
                     .nonNativeVisibleMenu
+            case "padded-notch":
+                    .nonNativePaddedNotch
             default:
                 defaultValue
             }
@@ -231,6 +250,17 @@ extension Ghostty {
             return String(cString: ptr)
         }
 
+        var macosWindowButtons: MacOSWindowButtons {
+            let defaultValue = MacOSWindowButtons.visible
+            guard let config = self.config else { return defaultValue }
+            var v: UnsafePointer<Int8>? = nil
+            let key = "macos-window-buttons"
+            guard ghostty_config_get(config, &v, key, UInt(key.count)) else { return defaultValue }
+            guard let ptr = v else { return defaultValue }
+            let str = String(cString: ptr)
+            return MacOSWindowButtons(rawValue: str) ?? defaultValue
+        }
+
         var macosTitlebarStyle: String {
             let defaultValue = "transparent"
             guard let config = self.config else { return defaultValue }
@@ -252,6 +282,17 @@ extension Ghostty {
             return MacOSTitlebarProxyIcon(rawValue: str) ?? defaultValue
         }
 
+        var macosDockDropBehavior: MacDockDropBehavior {
+            let defaultValue = MacDockDropBehavior.new_tab
+            guard let config = self.config else { return defaultValue }
+            var v: UnsafePointer<Int8>? = nil
+            let key = "macos-dock-drop-behavior"
+            guard ghostty_config_get(config, &v, key, UInt(key.count)) else { return defaultValue }
+            guard let ptr = v else { return defaultValue }
+            let str = String(cString: ptr)
+            return MacDockDropBehavior(rawValue: str) ?? defaultValue
+        }
+
         var macosWindowShadow: Bool {
             guard let config = self.config else { return false }
             var v = false;
@@ -271,6 +312,24 @@ extension Ghostty {
             return MacOSIcon(rawValue: str) ?? defaultValue
         }
 
+        var macosCustomIcon: String {
+            #if os(macOS)
+            let homeDirURL = FileManager.default.homeDirectoryForCurrentUser
+            let ghosttyConfigIconPath = homeDirURL.appendingPathComponent(
+                ".config/ghostty/Ghostty.icns",
+                conformingTo: .fileURL).path()
+            let defaultValue = ghosttyConfigIconPath
+            guard let config = self.config else { return defaultValue }
+            var v: UnsafePointer<Int8>? = nil
+            let key = "macos-custom-icon"
+            guard ghostty_config_get(config, &v, key, UInt(key.count)) else { return defaultValue }
+            guard let ptr = v else { return defaultValue }
+            return String(cString: ptr)
+            #else
+            return ""
+            #endif
+        }
+
         var macosIconFrame: MacOSIconFrame {
             let defaultValue = MacOSIconFrame.aluminum
             guard let config = self.config else { return defaultValue }
@@ -300,6 +359,16 @@ extension Ghostty {
             return buffer.map { .init(ghostty: $0) }
         }
 
+        var macosHidden: MacHidden {
+            guard let config = self.config else { return .never }
+            var v: UnsafePointer<Int8>? = nil
+            let key = "macos-hidden"
+            guard ghostty_config_get(config, &v, key, UInt(key.count)) else { return .never }
+            guard let ptr = v else { return .never }
+            let str = String(cString: ptr)
+            return MacHidden(rawValue: str) ?? .never
+        }
+
         var focusFollowsMouse : Bool {
             guard let config = self.config else { return false }
             var v = false;
@@ -435,6 +504,14 @@ extension Ghostty {
             let str = String(cString: ptr)
             return QuickTerminalSpaceBehavior(fromGhosttyConfig: str) ?? .move
         }
+
+        var quickTerminalSize: QuickTerminalSize {
+            guard let config = self.config else { return QuickTerminalSize() }
+            var v = ghostty_config_quick_terminal_size_s()
+            let key = "quick-terminal-size"
+            guard ghostty_config_get(config, &v, key, UInt(key.count)) else { return QuickTerminalSize() }
+            return QuickTerminalSize(from: v)
+        }
         #endif
 
         var resizeOverlay: ResizeOverlay {
@@ -466,6 +543,14 @@ extension Ghostty {
             return v;
         }
 
+        var undoTimeout: Duration {
+            guard let config = self.config else { return .seconds(5) }
+            var v: UInt = 0
+            let key = "undo-timeout"
+            _ = ghostty_config_get(config, &v, key, UInt(key.count))
+            return .milliseconds(v)
+        }
+
         var autoUpdate: AutoUpdate? {
             guard let config = self.config else { return nil }
             var v: UnsafePointer<Int8>? = nil
@@ -502,6 +587,25 @@ extension Ghostty {
             _ = ghostty_config_get(config, &v, key, UInt(key.count))
             return v
         }
+
+        var maximize: Bool {
+            guard let config = self.config else { return true }
+            var v = false;
+            let key = "maximize"
+            _ = ghostty_config_get(config, &v, key, UInt(key.count))
+            return v
+        }
+
+        var macosShortcuts: MacShortcuts {
+            let defaultValue = MacShortcuts.ask
+            guard let config = self.config else { return defaultValue }
+            var v: UnsafePointer<Int8>? = nil
+            let key = "macos-shortcuts"
+            guard ghostty_config_get(config, &v, key, UInt(key.count)) else { return defaultValue }
+            guard let ptr = v else { return defaultValue }
+            let str = String(cString: ptr)
+            return MacShortcuts(rawValue: str) ?? defaultValue
+        }
     }
 }
 
@@ -514,6 +618,31 @@ extension Ghostty.Config {
         case download
     }
 
+    struct BellFeatures: OptionSet {
+        let rawValue: CUnsignedInt
+
+        static let system = BellFeatures(rawValue: 1 << 0)
+        static let audio = BellFeatures(rawValue: 1 << 1)
+        static let attention = BellFeatures(rawValue: 1 << 2)
+        static let title = BellFeatures(rawValue: 1 << 3)
+    }
+    
+    enum MacDockDropBehavior: String {
+        case new_tab = "new-tab"
+        case new_window = "new-window"
+    }
+
+    enum MacHidden : String {
+        case never
+        case always
+    }
+
+    enum MacShortcuts: String {
+        case allow
+        case deny
+        case ask
+    }
+
     enum ResizeOverlay : String {
         case always
         case never
