@@ -3,7 +3,7 @@
 class Weak<T: AnyObject> {
     weak var value: T?
 
-    init(_ value: T) {
+    init(_ value: T? = nil) {
         self.value = value
     }
 }
